#pragma once

#include "config.h"
#include "service_proto.h"
#include "stream_proto.h"
#include <vector>
#include <memory>

#ifdef _WIN32
#include <winsock.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#define INVALID_SOCKET (-1)
typedef int SOCKET;
#endif

#if (defined _WIN32 && defined RF627_LIBRARY && !defined TEST)
#define RF627_API __declspec(dllexport)
#else
#define RF627_API
#endif

/*! Structure to store a point of profile
 */
struct rf627_point
{
    double x;
    double z;
};

/*! Structure to store a profile
 */
struct rf627_profile
{
    rf627_stream_msg_t header;
    std::vector<rf627_point> points;
    std::vector<uint16_t> pixels;
};

class rf627;
using rf627_list = std::vector<std::shared_ptr<rf627>>;

class RF627_API rf627
{
public:
    rf627() = delete;
    rf627(const rf627&) = delete;
    rf627(rf627_device_info_t* info, uint16_t init_msg_count = 1);
    virtual ~rf627();

    /*! Initialize library
     * Must be called once before further calls to any library functions
     */
    static bool init();
    /*! Cleanup resources allocated with init() function
     */
    static void cleanup();
    /*! Return SDK version
     */
    static uint32_t version();

    /*! Return const reference to structure containing device information
     */
    const rf627_device_info_t& info() const;

    /*! Search for RF627 devices over network
     * \param ok            :   ptr to boolean to indicate error
     * \return list of rf627 objects
     */
    static rf627_list search(bool *ok = nullptr);
    /*! Establish connection to the device
     * \return true on success
     */
    bool connect();
    /*! Close connection to the device
     */
    void disconnect();

    /*! Read parameters from device to internal structure. This structure is accessible via params() function
     * \param pparams         :   ptr to rf627_user_params_t structure; if not null parameters will be copied to this struct
     * \return true on success
     */
    bool read_params(rf627_user_params_t* pparams = nullptr);
    /*! Write parameters to device's memory
     * \param pparams         :   ptr to rf627_user_params_t structure; if not null parameters will be copied from this struct before writing to device
     * \return true on success
     */
    bool write_params(rf627_user_params_t* pparams = nullptr);
    /*! Save parameters from device's memory to flash
     * \return true on success
     */
    bool flush_params();
    /*! Reset parameters to factory set
     * \return true on success
     */
    bool reset_params();
    /*! Reboot the device
     * \return true on success
     */
    bool reboot();

    /*! Read specific block of parameters from device to internal structure. This structure is accessible as params() substructure
     * \param pparams         :   ptr to rf627_general_params_t structure; if not null parameters will be copied to this struct
     * \return true on success
     */
    bool read_general_params(rf627_general_params_t* pparams = nullptr);
    bool read_sysmon_params(rf627_sysmon_params_t* pparams = nullptr);
    bool read_rf625compat_params(rf627_rf625compat_params_t* pparams = nullptr);
    bool read_sensor_params(rf627_sensor_params_t* pparams = nullptr);
    bool read_roi_params(rf627_roi_params_t* pparams = nullptr);
    bool read_network_params(rf627_network_params_t* pparams = nullptr);
    bool read_stream_params(rf627_stream_params_t* pparams = nullptr);
    bool read_image_processing_params(rf627_image_processing_params_t* pparams = nullptr);
    bool read_laser_params(rf627_laser_params_t* pparams = nullptr);
    bool read_inputs_params(rf627_inputs_params_t* pparams = nullptr);
    bool read_outputs_params(rf627_outputs_params_t* pparams = nullptr);

    /*! Return ref to internal parameters structure
     */
    inline rf627_user_params_t &params() {return m_params;}

    /*! Get measurement from scanner's data stream
     * \param profile         :   ref to rf627_profile structure to be filled with measurement result
     * \return true on success
     */
    bool get_result(rf627_profile& profile);

    /*! Read image from CMOS (each pixel is 8-bit value representing brightess in range of 0..255
     * \param ppixmap         :   a ptr to buffer to fill with image pixels; must be RF627_IMAGE_SIZE bytes
     * \return true on success
     */
    bool get_image(uint8_t* ppixmap);

    /*! Write firmware image file to device's memory
     * \param file_name         :   name of file to transfer (normally *.rf627)
     * \return true on success
     */
    bool write_firmware_image(const char* file_name);
    /*! Save firmware image from device's memory to flash. Do not call this function unless write_firmware_image() returns true!
     * \return true on success
     */
    bool flush_firmware_image();

    /*! Read log entries from the device
     * \param nstart_line      :   number of start line
     * \param plog_entries     :   array of rf627_log_record_t structs to store log lines read; array items count must be equal to nlines
     * \param nlines           :   number of lines to read; must be less or equal to RF627_MAX_LOG_ENTRIES_PER_PAYLOAD
     * \return true on success
     */
    bool read_log(uint32_t nstart_line, rf627_log_record_t *plog_entries, int nlines);
    /*! Get total log entries count
     * \return number of log entries; 0 on error
     */
    uint32_t read_log_record_count();

    /*! Get ptr to string containing the last error message
     */
    const char *error_msg();

private:
    rf627_device_info_t m_info;
    rf627_user_params_t m_params;
    uint16_t m_msg_count;
    SOCKET m_svc_sock,
           m_data_sock;
};
